// Code by SoundOfSpouting#6980 (UID: 151149148639330304)

class Bit is HasEq[Bit]
	let a: Bool val
	new val create(a': Bool) =>
		a = a'
	new val from_unsigned_integer[A: UnsignedInteger[A] val](n: UnsignedInteger[A]) =>
		a = n != A.min_value()
	fun val to_unsigned_integer[A: UnsignedInteger[A] val](): A =>
		let one = A.max_value() / A.max_value()
		let zero = one - one
		if a then one else zero end
	fun eq(bit: Bit box^): Bool =>
		a == bit.a

class Byte is HasEq[Byte]
	let data: (Bit val, Bit val, Bit val, Bit val, Bit val, Bit val, Bit val, Bit val)
	new val create(a: Bit val, b: Bit val, c: Bit val, d: Bit val, e: Bit val, f: Bit val, g: Bit val, h: Bit val) =>
		data = (a, b, c, d, e, f, g, h)
	new val from_unsigned_integer[A: UnsignedInteger[A] val](n: UnsignedInteger[A]) =>
		let one = A.max_value() / A.max_value()
		let two = one + one
		let three = two + one
		let four = two * two
		let five = four + one
		let six = five + one
		let seven = six + one
		data = (
			Bit.from_unsigned_integer[A](n.shr(three).mod(two)),
			Bit.from_unsigned_integer[A](n.shr(one).mod(two)),
			Bit.from_unsigned_integer[A](n.shr(four).mod(two)),
			Bit.from_unsigned_integer[A](n.shr(five).mod(two)),
			Bit.from_unsigned_integer[A](n.shr(two).mod(two)),
			Bit.from_unsigned_integer[A](n.shr(six).mod(two)),
			Bit.from_unsigned_integer[A](n.shr(seven).mod(two)),
			Bit.from_unsigned_integer[A](n.mod(two))
		)
	fun to_unsigned_integer[A: UnsignedInteger[A] val](): A =>
		let one = A.max_value() / A.max_value()
		(let a, let b, let c, let d, let e, let f, let g, let h) = data
		g.to_unsigned_integer[A]().shl(one)
			.add(f.to_unsigned_integer[A]()).shl(one)
			.add(d.to_unsigned_integer[A]()).shl(one)
			.add(c.to_unsigned_integer[A]()).shl(one)
			.add(a.to_unsigned_integer[A]()).shl(one)
			.add(e.to_unsigned_integer[A]()).shl(one)
			.add(b.to_unsigned_integer[A]()).shl(one)
			.add(h.to_unsigned_integer[A]())
	fun eq(byte: Byte box^): Bool =>
		(data._1 == byte.data._1) and
		(data._2 == byte.data._2) and
		(data._3 == byte.data._3) and
		(data._4 == byte.data._4) and
		(data._5 == byte.data._5) and
		(data._6 == byte.data._6) and
		(data._7 == byte.data._7) and
		(data._8 == byte.data._8)

type StackNodeOptional[A] is (StackNode[A] | None)

class Stack[A]
	var _left: StackNodeOptional[A] = None
	var _right: StackNodeOptional[A] = None
	fun ref pushL(value: A) =>
		let prevLeft = _left = StackNode[A](None, _left, consume value)
		match prevLeft | let node: StackNode[A] => node.left = _left end
		if _right is None then _right = _left end
	fun ref pushR(value: A) =>
		let prevRight = _right = StackNode[A](_right, None, consume value)
		match prevRight | let node: StackNode[A] => node.right = _right end
		if _left is None then _left = _right end
	fun ref popL(): A ? =>
		if _right is _left then _right = None end
		let leftNode = _left as StackNode[A]
		_left = leftNode.right
		leftNode.value
	fun ref popR(): A ? =>
		if _left is _right then _left = None end
		let rightNode = _right as StackNode[A]
		_right = rightNode.left
		rightNode.value
	fun peekL(): this->A ? =>
		(_left as this->StackNode[A]).value
	fun peekR(): this->A ? =>
		(_right as this->StackNode[A]).value

class StackNode[A]
	var left: StackNodeOptional[A]
	var right: StackNodeOptional[A]
	let value: A
	new create(left': StackNodeOptional[A], right': StackNodeOptional[A], value': A) =>
		left = left'
		right = right'
		value = consume value'

type TreeNode[A: HasEq[A] val] is (TreeBranchNode[A] | TreeLeafNode[A])

class Tree[A: HasEq[A] val]
	let head: TreeNode[A]
	new iso parsePostfix(delimeter: A, input: Array[A] val) ? =>
		let stack: Stack[TreeNode[A]] ref = Stack[TreeNode[A]]
		for value in input.values() do
			if value == delimeter then
				let right: TreeNode[A] = stack.popR()?
				let left : TreeNode[A] = stack.popR()?
				stack.pushR(TreeBranchNode[A](left, right))
			else stack.pushR(TreeLeafNode[A](value)) end
		end
		head = stack.peekR()?
		head
	fun ref invert() =>
		try (head as TreeBranchNode[A]).invert() end
	fun serializePrefix(delimeter: A): Iterator[A] =>
		head.serializePrefix(delimeter)

class TreeBranchNode[A: HasEq[A] val]
	var left: TreeNode[A]
	var right: TreeNode[A]
	new create(left': TreeNode[A], right': TreeNode[A]) =>
		left = left'
		right = right'
	fun ref invert() =>
		left = right = left
		try (left as TreeBranchNode[A] ref).invert() end
		try (right as TreeBranchNode[A] ref).invert() end
	fun serializePrefix(delimeter: A): Iterator[A] =>
		let stack: Stack[Iterator[A]] ref = Stack[Iterator[A]]
		stack.pushR(SingletonIterator[A](delimeter))
		stack.pushR(left.serializePrefix(delimeter))
		stack.pushR(right.serializePrefix(delimeter))
		ConcatIterators[A](stack)

class TreeLeafNode[A: HasEq[A] val]
	let value: A
	new create(value': A) =>
		value = consume value'
	fun serializePrefix(delimeter: A): Iterator[A] =>
		SingletonIterator[A](value)

primitive IteratorToArray[A: Any val]
	fun apply(a: Iterator[A]): Array[A] =>
		Array[A].>concat(a)

trait Map[A: Any val, B: Any val]
	fun apply(a: A): B

primitive UnsignedIntegerToByte[A: UnsignedInteger[A] val] is Map[A, Byte val]
	fun apply(a: A): Byte val =>
		Byte.from_unsigned_integer[A](a)

primitive ByteToUnsignedInteger[A: UnsignedInteger[A] val] is Map[Byte val, A]
	fun apply(b: Byte val): A =>
		b.to_unsigned_integer[A]()

primitive IdentityMap[A: Any val] is Map[A, A]
	fun apply(a: A): A =>
		a

class MapIterator[A: Any val, B: Any val] is Iterator[B]
	let itr: Iterator[A]
	let map: Map[A, B] val
	new create(itr': Iterator[A], map': Map[A, B] val) =>
		itr = consume itr'
		map = consume map'
	fun ref has_next(): Bool =>
		itr.has_next()
	fun ref next(): B ? =>
		map(itr.next()?)

primitive MapArray[A: Any val, B: Any val]
	fun apply(a: Array[A] val, map: Map[A, B] val): Array[B] =>
		IteratorToArray[B](MapIterator[A, B](a.values(), map))

class ConcatIterators[A] is Iterator[A]
	let stack: Stack[Iterator[A]]
	new create(stack': Stack[Iterator[A]]) =>
		stack = stack'
	fun ref has_next(): Bool =>
		try
			let itr = stack.peekL()?
			if itr.has_next() then true
			else
				stack.popL()?
				has_next()
			end
		else false end
	fun ref next(): A ? =>
		stack.peekL()?.next()?

class SingletonIterator[A] is Iterator[A]
	let value: A
	var unused: Bool = true
	new create(value': A) =>
		value = consume value'
	fun ref has_next(): Bool =>
		unused
	fun ref next(): A =>
		unused = false
		value

class Main
	fun entry(input: String): String =>
		try
			let tree = Tree[Byte val].parsePostfix(Byte.from_unsigned_integer[U8](','), recover val IteratorToArray[Byte val](MapIterator[U8, Byte val](input.values(), UnsignedIntegerToByte[U8])) end)?
			tree.invert()
			String.from_array(recover val IteratorToArray[U8](MapIterator[Byte val, U8]((consume tree).serializePrefix(Byte.from_unsigned_integer[U8]('`')), ByteToUnsignedInteger[U8])) end)
		else input end
